# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.ops.build_job import BuildSchema
from polyaxon_schemas.ops.operation import BaseOpConfig, BaseOpSchema


class BaseRunSchema(BaseOpSchema):
    build = fields.Nested(BuildSchema)

    @staticmethod
    def schema_config():
        return BaseRunConfig


class BaseRunConfig(BaseOpConfig):
    SCHEMA = BaseRunSchema
    IDENTIFIER = 'run'
    REDUCED_ATTRIBUTES = BaseOpConfig.REDUCED_ATTRIBUTES + ['build']

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None,
                 build=None,
                 ):
        super(BaseRunConfig, self).__init__(
            version=version,
            kind=kind,
            logging=logging,
            name=name,
            description=description,
            tags=tags,
            environment=environment,
            params=params,
            declarations=declarations,
            inputs=inputs,
            outputs=outputs,
        )
        self.build = build



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.layers.noise import (
    AlphaDropoutConfig,
    GaussianDropoutConfig,
    GaussianNoiseConfig
)


class TestNoiseConfigs(TestCase):
    def test_gaussian_noise_config(self):
        config_dict = {
            'stddev': 1.0
        }

        config = GaussianNoiseConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_gaussian_dropout_config(self):
        config_dict = {
            'rate': 0.8
        }

        config = GaussianDropoutConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_alpha_dropout_config(self):
        config_dict = {
            'rate': 0.8,
            'noise_shape': [1, 1],
            'seed': None
        }

        config = AlphaDropoutConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

from unittest import TestCase

from polyaxon_schemas.exceptions import PolyaxonSchemaError
from polyaxon_schemas.polyaxonfile import PolyaxonFile


class TestPolyaxonfileWithPipelines(TestCase):
    def test_pipeline_with_no_ops_raises(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/pipeline_with_no_ops.yml'))
        with self.assertRaises(PolyaxonSchemaError):
            plx_file.specification.parse_data()

    def test_pipeline_with_no_templates_raises(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/pipeline_with_no_templates.yml'))
        with self.assertRaises(PolyaxonSchemaError):
            plx_file.specification.parse_data()

    def test_pipeline_ops_not_corresponding_to_templates(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/pipeline_ops_not_corresponding_to_templates.yml'))
        with self.assertRaises(PolyaxonSchemaError):
            plx_file.specification.parse_data()

    def test_cyclic_pipeline_raises(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/cyclic_pipeline.yml'))
        with self.assertRaises(PolyaxonSchemaError):
            plx_file.specification.parse_data()

    def test_cron_pipeline(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/simple_cron_pipeline.yml'))
        spec = plx_file.specification
        spec.parse_data()
        assert len(spec.config.ops) == 1
        assert spec.config.ops[0].name == 'cron-task'
        assert spec.config.concurrency is None
        assert spec.config.schedule is not None
        assert spec.config.schedule.kind == 'cron'
        assert spec.config.schedule.cron == '0 0 * * *'

    def test_interval_pipeline(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/simple_recurrent_pipeline.yml'))
        spec = plx_file.specification
        spec.parse_data()
        assert len(spec.config.ops) == 1
        assert spec.config.ops[0].name == 'recurrent-task'
        assert spec.config.concurrency is None
        assert spec.config.schedule is not None
        assert spec.config.schedule.kind == 'interval'
        assert spec.config.schedule.start_at.year == 2019
        assert spec.config.schedule.frequency == 120
        assert spec.config.schedule.depends_on_past is True

    def test_sequential_pipeline(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/simple_sequential_pipeline.yml'))
        spec = plx_file.specification
        spec.parse_data()
        assert len(spec.config.ops) == 4
        assert spec.config.ops[0].name == 'job1'
        assert spec.config.ops[1].name == 'job2'
        assert spec.config.ops[1].dependencies == ['job1']
        assert spec.config.ops[2].name == 'experiment1'
        assert spec.config.ops[2].dependencies == ['job2']
        assert spec.config.ops[3].name == 'experiment2'
        assert spec.config.ops[3].dependencies == ['experiment1']
        assert spec.config.sort_topologically(spec.config.dag) == [
            ['job1'], ['job2'], ['experiment1'], ['experiment2']]
        assert spec.config.concurrency is None
        assert spec.config.schedule is None

    def test_parallel_pipeline(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/simple_parallel_pipeline.yml'))
        spec = plx_file.specification
        spec.parse_data()
        assert len(spec.config.ops) == 4
        assert spec.config.ops[0].name == 'job1'
        assert spec.config.ops[0].dependencies is None
        assert spec.config.ops[1].name == 'job2'
        assert spec.config.ops[1].dependencies is None
        assert spec.config.ops[2].name == 'experiment1'
        assert spec.config.ops[2].dependencies is None
        assert spec.config.ops[3].name == 'experiment2'
        assert spec.config.ops[3].dependencies is None
        assert set(spec.config.sort_topologically(spec.config.dag)[0]) == {
            'job1', 'job2', 'experiment1', 'experiment2'}
        assert spec.config.concurrency == 2
        assert spec.config.schedule is None

    def test_dag_pipeline(self):
        plx_file = PolyaxonFile(os.path.abspath(
            'tests/fixtures/pipelines/simple_dag_pipeline.yml'))
        spec = plx_file.specification
        spec.parse_data()
        assert len(spec.config.ops) == 5
        assert spec.config.ops[0].name == 'job1'
        assert spec.config.ops[1].name == 'experiment1'
        assert spec.config.ops[1].dependencies == ['job1']
        assert spec.config.ops[2].name == 'experiment2'
        assert spec.config.ops[2].dependencies == ['job1']
        assert spec.config.ops[3].name == 'experiment3'
        assert spec.config.ops[3].dependencies == ['job1']
        assert spec.config.ops[4].name == 'job2'
        assert spec.config.ops[4].dependencies == ['experiment1', 'experiment2', 'experiment3']
        srorted_dag = spec.config.sort_topologically(spec.config.dag)
        assert srorted_dag[0] == ['job1']
        assert set(srorted_dag[1]) == {'experiment1', 'experiment2', 'experiment3'}
        assert srorted_dag[2] == ['job2']
        assert spec.config.concurrency == 3
        assert spec.config.schedule is None



class TriggerPolicy(object):
    ALL_SUCCEEDED = 'all_succeeded'
    ALL_FAILED = 'all_failed'
    ALL_DONE = 'all_done'
    ONE_SUCCEEDED = 'one_succeeded'
    ONE_FAILED = 'one_failed'
    ONE_DONE = 'one_done'

    VALUES = {ALL_SUCCEEDED, ALL_FAILED, ALL_DONE, ONE_SUCCEEDED, ONE_FAILED, ONE_DONE}
    CHOICES = (
        (ALL_SUCCEEDED, ALL_SUCCEEDED),
        (ALL_FAILED, ALL_FAILED),
        (ALL_DONE, ALL_DONE),
        (ONE_SUCCEEDED, ONE_SUCCEEDED),
        (ONE_FAILED, ONE_FAILED),
        (ONE_DONE, ONE_DONE),
    )


class StatusTriggerPolicy(object):
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    DONE = 'done'

    VALUES = {SUCCEEDED, FAILED, DONE}
    CHOICES = (
        (SUCCEEDED, SUCCEEDED),
        (FAILED, FAILED),
        (DONE, DONE),
    )


class ExpressionTriggerPolicy(object):
    EQ = 'eq'
    NE = 'ne'
    IS_NULL = 'isnull'
    NOT_NULL = 'not_null'

    VALUES = {EQ, NE}
    CHOICES = (
        (EQ, EQ),
        (NE, NE),
        (IS_NULL, IS_NULL),
        (NOT_NULL, NOT_NULL),
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.initializations import (
    GlorotUniformInitializerConfig,
    OrthogonalInitializerConfig,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.recurrent import (
    GRUConfig,
    LSTMConfig,
    RecurrentConfig,
    SimpleRNNConfig
)


class TestRecurrentConfigs(TestCase):
    def test_recurrent_config(self):
        config_dict = {
            'return_sequences': False,
            'return_state': False,
            'go_backwards': False,
            'stateful': False,
            'unroll': False,
            'implementation': 0,
        }
        config = RecurrentConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_simple_rnn_config(self):
        config_dict = {
            'units': 3,
            'activation': 'tanh',
            'use_bias': True,
            'kernel_initializer': GlorotUniformInitializerConfig().to_schema(),
            'recurrent_initializer': OrthogonalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'kernel_regularizer': None,
            'recurrent_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'recurrent_constraint': None,
            'bias_constraint': None,
            'dropout': 0.,
            'recurrent_dropout': 0.,
        }
        config = SimpleRNNConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_gru_config(self):
        config_dict = {
            'units': 3,
            'activation': 'tanh',
            'recurrent_activation': 'hard_sigmoid',
            'use_bias': True,
            'kernel_initializer': GlorotUniformInitializerConfig().to_schema(),
            'recurrent_initializer': OrthogonalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'kernel_regularizer': None,
            'recurrent_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'recurrent_constraint': None,
            'bias_constraint': None,
            'dropout': 0.,
            'recurrent_dropout': 0.,
        }
        config = GRUConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_lstm_config(self):
        config_dict = {
            'units': 3,
            'activation': 'tanh',
            'recurrent_activation': 'hard_sigmoid',
            'use_bias': True,
            'kernel_initializer': GlorotUniformInitializerConfig().to_schema(),
            'recurrent_initializer': OrthogonalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'unit_forget_bias': True,
            'kernel_regularizer': None,
            'recurrent_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'recurrent_constraint': None,
            'bias_constraint': None,
            'dropout': 0.,
            'recurrent_dropout': 0.,
        }
        config = LSTMConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validate, validates_schema
from rhea import RheaError, parser

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.io.types import IOTypes


def validate_io(name, iotype, default, is_optional, is_list, is_flag):
    if iotype and default:
        try:
            parser.TYPE_MAPPING[iotype](
                key=name,
                value=default,
                is_list=is_list,
                is_optional=is_optional,
                default=None,
                options=None)

        except RheaError as e:
            raise ValidationError('IO `%s` Could not parse default value `%s`, '
                                  'an error was encountered: %s' % (name, default, e))

    if not is_optional and default:
        raise ValidationError('IO `%s` is not optional and has default value `%s`. '
                              'Please either make it optional or remove the default value.')

    if is_flag and iotype != IOTypes.BOOL:
        raise ValidationError('IO type `{}` cannot be a flag, iut must be a `{}`'.format(
            iotype, IOTypes.BOOL
        ))


class IOSchema(BaseSchema):
    name = fields.Str()
    description = fields.Str(allow_none=True)
    iotype = fields.Str(allow_none=True,
                        data_key="type",
                        validate=validate.OneOf(IOTypes.VALUES))
    default = fields.Raw(allow_none=True)
    is_optional = fields.Bool(allow_none=True)
    is_list = fields.Bool(allow_none=True)
    is_flag = fields.Bool(allow_none=True)
    options = fields.List(fields.Raw(), allow_none=True)

    @staticmethod
    def schema_config():
        return IOConfig

    @validates_schema
    def validate_io(self, values):
        validate_io(name=values.get('name'),
                    iotype=values.get('iotype'),
                    default=values.get('default'),
                    is_list=values.get('is_list'),
                    is_optional=values.get('is_optional'),
                    is_flag=values.get('is_flag'))


class IOConfig(BaseConfig):
    SCHEMA = IOSchema
    IDENTIFIER = 'io'
    REDUCED_ATTRIBUTES = [
        'description',
        'type',
        'default',
        'is_optional',
        'is_flag',
        'is_list',
        'options']

    def __init__(self,
                 name,
                 description=None,
                 iotype=None,
                 default=None,
                 is_optional=None,
                 is_list=None,
                 is_flag=None,
                 options=None):
        validate_io(name=name,
                    iotype=iotype,
                    default=default,
                    is_optional=is_optional,
                    is_list=is_list,
                    is_flag=is_flag)

        self.name = name
        self.description = description
        self.iotype = iotype
        self.default = default
        self.is_optional = is_optional
        self.is_list = is_list
        self.is_flag = is_flag
        self.options = options

    def validate_value(self, value):
        if self.iotype is None:
            return value

        try:
            return parser.TYPE_MAPPING[self.iotype](
                key=self.name,
                value=value,
                is_list=self.is_list,
                is_optional=self.is_optional,
                default=self.default,
                options=self.options)
        except RheaError as e:
            raise ValidationError('Could not parse value `%s`, '
                                  'an error was encountered: %s' % (value, e))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import copy

from hestia.cached_property import cached_property

from polyaxon_schemas.polyflow.pipeline import PipelineConfig
from polyaxon_schemas.specs import kinds
from polyaxon_schemas.specs.base import BaseSpecification


class PipelineSpecification(BaseSpecification):
    """The polyaxonfile specification for pipelines.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = kinds.PIPELINE

    TEMPLATES = 'templates'
    OPS = 'ops'
    SCHEDULE = 'schedule'
    CONCURRENCY = 'concurrency'

    SECTIONS = BaseSpecification.SECTIONS + (TEMPLATES, OPS, SCHEDULE, CONCURRENCY)

    OP_PARSING_SECTIONS = BaseSpecification.OP_PARSING_SECTIONS + (
        TEMPLATES, OPS, SCHEDULE, CONCURRENCY)

    HEADER_SECTIONS = BaseSpecification.HEADER_SECTIONS + (BaseSpecification.BACKEND,)

    POSSIBLE_SECTIONS = BaseSpecification.POSSIBLE_SECTIONS + (
        BaseSpecification.BACKEND, BaseSpecification.ENVIRONMENT,
        TEMPLATES, OPS, SCHEDULE, CONCURRENCY
    )

    CONFIG = PipelineConfig

    @cached_property
    def backend(self):
        return self.raw_config.backend

    def _get_config(self, data):
        config = self.CONFIG.from_dict(copy.deepcopy(data))
        return config

    def parse_data(self, context=None):
        self._config_data.process_dag()
        self._config_data.validate_dag()
        self._config_data.process_templates()
        self._config = self._config_data
        return self



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ops.experiment import ExperimentConfig, ExperimentSchema
from polyaxon_schemas.ops.group.hptuning import HPTuningSchema


class GroupSchema(ExperimentSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('group'))
    hptuning = fields.Nested(HPTuningSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return GroupConfig


class GroupConfig(ExperimentConfig):
    SCHEMA = GroupSchema
    IDENTIFIER = 'group'
    REDUCED_ATTRIBUTES = ExperimentConfig.REDUCED_ATTRIBUTES + ['hptuning']

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None,
                 build=None,
                 backend=None,
                 framework=None,
                 run=None,
                 hptuning=None,
                 ):
        super(GroupConfig, self).__init__(
            version=version,
            kind=kind,
            logging=logging,
            name=name,
            description=description,
            tags=tags,
            environment=environment,
            params=params,
            declarations=declarations,
            inputs=inputs,
            outputs=outputs,
            build=build,
            backend=backend,
            framework=framework,
            run=run,
        )
        self.hptuning = hptuning



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import uuid

from unittest import TestCase

from hestia.tz_utils import local_now

from polyaxon_schemas.api.data import DataConfig, DataDetailsConfig, DatasetConfig


class TestDataConfigs(TestCase):
    def test_data_details_config(self):
        config_dict = {
            'state': 'state',
            'size': 1.4,
            'uri': 'http://www.foo.com/data'
        }
        config = DataDetailsConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_data_config(self):
        config_dict = {
            'uuid': uuid.uuid4().hex,
            'name': 'foo',
            'created_at': local_now().isoformat(),
            'description': 'foo data',
            'details': DataDetailsConfig(
                state='state', size=1.4, uri='http://www.foo.com/data').to_dict(),
            'version': None,
            'resource_id': '1'
        }
        config = DataConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

    def test_dataset_config(self):
        config_dict = {
            'uuid': uuid.uuid4().hex,
            'name': 'foo',
            'description': 'foo data',
            'is_public': True
        }
        config = DatasetConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema


class BaseExplorationSchema(BaseSchema):
    is_continuous = fields.Bool(default=False, missing=False)

    @staticmethod
    def schema_config():
        return BaseExplorationConfig


class BaseExplorationConfig(BaseConfig):
    def __init__(self, is_continuous=False):
        self.is_continuous = is_continuous


class ConstantExplorationSchema(BaseExplorationSchema):
    value = fields.Float(default=0.5, missing=0.5)

    @staticmethod
    def schema_config():
        return ConstantExplorationConfig


class ConstantExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'Constant'
    SCHEMA = ConstantExplorationSchema

    def __init__(self, value=0.5, is_continuous=False):
        self.value = value
        super(ConstantExplorationConfig, self).__init__(is_continuous)


class GreedyExplorationSchema(BaseExplorationSchema):

    @staticmethod
    def schema_config():
        return GreedyExplorationConfig


class GreedyExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'Greedy'
    SCHEMA = GreedyExplorationSchema


class RandomExplorationSchema(BaseExplorationSchema):

    @staticmethod
    def schema_config():
        return RandomExplorationConfig


class RandomExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'Random'
    SCHEMA = RandomExplorationSchema


class DecayExplorationSchema(BaseExplorationSchema):
    exploration_rate = fields.Float(allow_none=True)
    decay_type = fields.Str(allow_none=True, validate=validate.OneOf(
        ['exponential_decay', 'inverse_time_decay', 'natural_exp_decay', 'piecewise_constant',
         'polynomial_decay']))
    start_decay_at = fields.Int(allow_none=True)
    stop_decay_at = fields.Int(allow_none=True)
    decay_rate = fields.Float(allow_none=True)
    staircase = fields.Bool(allow_none=True)
    decay_steps = fields.Int(allow_none=True)
    min_exploration_rate = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return DecayExplorationConfig


class DecayExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'Decay'
    SCHEMA = DecayExplorationSchema

    def __init__(self,
                 is_continuous=False,
                 exploration_rate=0.15,
                 decay_type='polynomial_decay',
                 start_decay_at=0,
                 stop_decay_at=1e9,
                 decay_rate=0.,
                 staircase=False,
                 decay_steps=100000,
                 min_exploration_rate=0):
        self.exploration_rate = exploration_rate
        self.decay_type = decay_type
        self.start_decay_at = start_decay_at
        self.stop_decay_at = stop_decay_at
        self.decay_rate = decay_rate
        self.staircase = staircase
        self.decay_steps = decay_steps
        self.min_exploration_rate = min_exploration_rate
        super(DecayExplorationConfig, self).__init__(is_continuous)


class RandomDecayExplorationSchema(BaseExplorationSchema):
    num_actions = fields.Int(allow_none=True)
    decay_type = fields.Str(allow_none=True, validate=validate.OneOf(
        ['exponential_decay', 'inverse_time_decay', 'natural_exp_decay', 'piecewise_constant',
         'polynomial_decay']))
    start_decay_at = fields.Int(allow_none=True)
    stop_decay_at = fields.Int(allow_none=True)
    decay_rate = fields.Float(allow_none=True)
    staircase = fields.Bool(allow_none=True)
    decay_steps = fields.Int(allow_none=True)
    min_exploration_rate = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return RandomDecayExplorationConfig


class RandomDecayExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'RandomDecay'
    SCHEMA = RandomDecayExplorationSchema

    def __init__(self,
                 is_continuous=False,
                 num_actions=None,
                 decay_type='polynomial_decay',
                 start_decay_at=0,
                 stop_decay_at=1e9,
                 decay_rate=0.,
                 staircase=False,
                 decay_steps=10000,
                 min_exploration_rate=0):
        self.num_actions = num_actions
        self.decay_type = decay_type
        self.start_decay_at = start_decay_at
        self.stop_decay_at = stop_decay_at
        self.decay_rate = decay_rate
        self.staircase = staircase
        self.decay_steps = decay_steps
        self.min_exploration_rate = min_exploration_rate
        super(RandomDecayExplorationConfig, self).__init__(is_continuous)


class OrnsteinUhlenbeckExplorationSchema(BaseExplorationSchema):
    num_actions = fields.Int(allow_none=True)
    sigma = fields.Float(allow_none=True)
    mu = fields.Float(allow_none=True)
    theta = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return OrnsteinUhlenbeckExplorationConfig


class OrnsteinUhlenbeckExplorationConfig(BaseExplorationConfig):
    IDENTIFIER = 'OrnsteinUhlenbeck'
    SCHEMA = OrnsteinUhlenbeckExplorationSchema

    def __init__(self, num_actions, sigma=0.3, mu=0, theta=0.15, **kwargs):
        self.num_actions = num_actions
        self.sigma = sigma
        self.mu = mu
        self.theta = theta
        super(OrnsteinUhlenbeckExplorationConfig, self).__init__(True)


class RegularizerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'exploration'
    __configs__ = {
        ConstantExplorationConfig.IDENTIFIER: ConstantExplorationConfig,
        GreedyExplorationConfig.IDENTIFIER: GreedyExplorationConfig,
        RandomExplorationConfig.IDENTIFIER: RandomExplorationConfig,
        DecayExplorationConfig.IDENTIFIER: DecayExplorationConfig,
        RandomDecayExplorationConfig.IDENTIFIER: RandomDecayExplorationConfig,
        OrnsteinUhlenbeckExplorationConfig.IDENTIFIER: OrnsteinUhlenbeckExplorationConfig
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from hestia.cached_property import cached_property
from marshmallow import EXCLUDE

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.ops.experiment import ExperimentConfig
from polyaxon_schemas.ops.experiment.environment import ExperimentEnvironmentConfig
from polyaxon_schemas.ops.run import RunConfig
from polyaxon_schemas.specs import kinds
from polyaxon_schemas.specs.base import BaseRunSpecification
from polyaxon_schemas.specs.frameworks import (
    HorovodSpecification,
    MPISpecification,
    MXNetSpecification,
    PytorchSpecification,
    TensorflowSpecification
)
from polyaxon_schemas.utils import TaskType


class ExperimentSpecification(BaseRunSpecification):
    """The Base polyaxonfile specification (parsing and validation of Polyaxonfiles/Configurations).

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        PARAMS: variables/modules that can be reused.
        BUILD: defines the build step where the user can set a docker image definition
        RUN: defines the run step where the user can run a command
    """
    _SPEC_KIND = kinds.EXPERIMENT  # pylint:disable=protected-access

    SECTIONS = BaseRunSpecification.SECTIONS

    HEADER_SECTIONS = BaseRunSpecification.HEADER_SECTIONS + (
        BaseRunSpecification.FRAMEWORK,
    )

    POSSIBLE_SECTIONS = BaseRunSpecification.POSSIBLE_SECTIONS + (
        BaseRunSpecification.FRAMEWORK,
        BaseRunSpecification.DECLARATIONS,
        BaseRunSpecification.PARAMS,
        BaseRunSpecification.RUN,
    )

    ENVIRONMENT_CONFIG = ExperimentEnvironmentConfig
    CONFIG = ExperimentConfig

    @cached_property
    def params(self):
        return self.parsed_data.get(self.PARAMS, None)

    @cached_property
    def backend(self):
        return self.config.backend

    @cached_property
    def framework(self):
        return self.config.framework

    @cached_property
    def run(self):
        return self.config.run

    @cached_property
    def cluster_def(self):
        cluster = {
            TaskType.MASTER: 1,
        }
        is_distributed = False
        environment = self.environment

        if not environment:
            return cluster, is_distributed

        if self.config.tensorflow:
            return TensorflowSpecification.get_cluster_def(
                cluster=cluster,
                framework_config=self.config.tensorflow)
        if self.config.horovod:
            return HorovodSpecification.get_cluster_def(
                cluster=cluster,
                framework_config=self.config.horovod)
        if self.config.mxnet:
            return MXNetSpecification.get_cluster_def(
                cluster=cluster,
                framework_config=self.config.mxnet)
        if self.config.pytorch:
            return PytorchSpecification.get_cluster_def(
                cluster=cluster,
                framework_config=self.config.pytorch)
        if self.config.mpi:
            return MPISpecification.get_cluster_def(
                cluster={},
                framework_config=self.config.mpi)

        # No specified framework, It should return default standalone mode cluster definition
        return cluster, is_distributed

    @cached_property
    def total_resources(self):
        environment = self.environment

        if not environment:
            return None

        cluster, is_distributed = self.cluster_def

        # Check if any framework is defined
        if self.config.tensorflow:
            return TensorflowSpecification.get_total_resources(
                master_resources=self.master_resources,
                environment=self.config.tensorflow,
                cluster=cluster,
                is_distributed=is_distributed
            )

        if self.config.horovod:
            return HorovodSpecification.get_total_resources(
                master_resources=self.master_resources,
                environment=self.config.horovod,
                cluster=cluster,
                is_distributed=is_distributed
            )

        if self.config.mxnet:
            return MXNetSpecification.get_total_resources(
                master_resources=self.master_resources,
                environment=self.config.mxnet,
                cluster=cluster,
                is_distributed=is_distributed
            )

        if self.config.pytorch:
            return PytorchSpecification.get_total_resources(
                master_resources=self.master_resources,
                environment=self.config.pytorch,
                cluster=cluster,
                is_distributed=is_distributed
            )

        if self.config.mpi:
            return MPISpecification.get_total_resources(
                master_resources=self.master_resources,
                environment=self.config.mpi,
                cluster=cluster,
                is_distributed=is_distributed
            )

        # default value is the master resources
        return self.master_resources

    @cached_property
    def master_resources(self):
        return self.environment.resources if self.environment else None

    @cached_property
    def master_node_selector(self):
        return self.environment.node_selector if self.environment else None

    @cached_property
    def master_affinity(self):
        return self.environment.affinity if self.environment else None

    @cached_property
    def master_tolerations(self):
        return self.environment.tolerations if self.environment else None

    @classmethod
    def create_specification(cls,  # pylint:disable=arguments-differ
                             build_config,
                             run_config,
                             to_dict=True):
        try:
            specification = BaseRunSpecification.create_specification(
                build_config=build_config, to_dict=True)
        except PolyaxonConfigurationError:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of BuildConfig.')

        if isinstance(run_config, RunConfig):
            r_config = run_config.to_light_dict()
        elif isinstance(run_config, Mapping):
            r_config = RunConfig.from_dict(run_config, unknown=EXCLUDE)
            r_config = r_config.to_light_dict()
        else:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of RunConfig.')

        specification[cls.KIND] = cls._SPEC_KIND
        specification[cls.RUN] = r_config

        if to_dict:
            return specification
        return cls.read(specification)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_graphs

from polyaxon_schemas.ml.graph import GraphConfig


class TestGraphConfigs(TestCase):
    def test_graph_config(self):
        config_dict = {
            'input_layers': ['image'],
            'output_layers': [['dense_0', 0, 0]],
            'layers': [
                {
                    'Conv2D': {
                        'filters': 64,
                        'strides': [1, 1],
                        'kernel_size': [2, 2],
                        'activation': 'relu',
                        'name': 'convolution_1',
                    }
                },
                {'Dense': {'units': 17, 'name': 'dense_0'}}
            ]
        }
        config = GraphConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert_equal_graphs(config_dict, config_to_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from hestia.tz_utils import local_now
from marshmallow import ValidationError

from polyaxon_schemas.polyflow.schedule import (
    CronScheduleConfig,
    IntervalScheduleConfig,
    ScheduleSchema
)


class TestScheduleConfigs(TestCase):
    def test_interval_schedule(self):
        config_dict = {'frequency': 2, 'start_at': 'foo'}
        with self.assertRaises(ValidationError):
            IntervalScheduleConfig.from_dict(config_dict)

        config_dict = {'frequency': 'foo', 'start_at': local_now().isoformat()}
        with self.assertRaises(ValidationError):
            IntervalScheduleConfig.from_dict(config_dict)

        config_dict = {'kind': 'cron',
                       'frequency': 2,
                       'start_at': local_now().isoformat(),
                       'end_at': local_now().isoformat()}
        with self.assertRaises(ValidationError):
            IntervalScheduleConfig.from_dict(config_dict)

        config_dict = {
            'frequency': 2,
            'start_at': local_now().isoformat()
        }
        IntervalScheduleConfig.from_dict(config_dict)

        config_dict = {
            'frequency': 2,
            'start_at': local_now().isoformat(),
            'end_at': local_now().isoformat()
        }
        IntervalScheduleConfig.from_dict(config_dict)

        config_dict = {
            'kind': 'interval',
            'frequency': 2,
            'start_at': local_now().isoformat(),
            'end_at': local_now().isoformat(),
            'depends_on_past': False,
        }
        IntervalScheduleConfig.from_dict(config_dict)

    def test_cron_schedule(self):
        config_dict = {'cron': 2, 'start_at': 'foo'}
        with self.assertRaises(ValidationError):
            CronScheduleConfig.from_dict(config_dict)

        config_dict = {'kind': 'interval',
                       'cron': '0 0 * * *',
                       'start_at': local_now().isoformat(),
                       'end_at': local_now().isoformat()}
        with self.assertRaises(ValidationError):
            CronScheduleConfig.from_dict(config_dict)

        config_dict = {
            'cron': '0 0 * * *',
        }
        CronScheduleConfig.from_dict(config_dict)

        config_dict = {
            'cron': '0 0 * * *',
            'start_at': local_now().isoformat(),
            'end_at': local_now().isoformat()
        }
        CronScheduleConfig.from_dict(config_dict)

        config_dict = {
            'kind': 'cron',
            'cron': '0 0 * * *',
            'start_at': local_now().isoformat(),
            'end_at': local_now().isoformat(),
            'depends_on_past': False,
        }
        CronScheduleConfig.from_dict(config_dict)

    def test_schedule(self):
        configs = [
            {
                'kind': 'interval',
                'frequency': 2,
                'start_at': local_now().isoformat(),
                'end_at': local_now().isoformat(),
                'depends_on_past': False,
            },
            {
                'kind': 'cron',
                'cron': '0 0 * * *',
                'start_at': local_now().isoformat(),
                'end_at': local_now().isoformat(),
                'depends_on_past': False,
            }
        ]

        ScheduleSchema().load(configs, many=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


class ExperimentFramework(object):
    TENSORFLOW = 'tensorflow'
    MXNET = 'mxnet'
    HOROVOD = 'horovod'
    PYTORCH = 'pytorch'

    VALUES = [TENSORFLOW, MXNET, HOROVOD, PYTORCH]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.environments.pods import EnvironmentConfig, EnvironmentSchema


class TensorflowClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)
    ps = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowClusterConfig


class TensorflowClusterConfig(BaseConfig):
    """
    Tensorflow cluster config.

    Args:
        master: list(str). The list of master host in a tensorflow cluster.
        worker: list(str). The list of worker host in a tensorflow cluster.
        ps: list(str). The list of ps host in a tensorflow cluster.
    """
    IDENTIFIER = 'tensorflow_cluster'
    SCHEMA = TensorflowClusterSchema

    def __init__(self, master=None, worker=None, ps=None):
        self.master = master
        self.worker = worker
        self.ps = ps


class HorovodClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return HorovodClusterConfig


class HorovodClusterConfig(BaseConfig):
    """
    Horovod cluster config.

    Args:
        master: list(str). The list of master host in a Horovod cluster.
        worker: list(str). The list of worker host in a Horovod cluster.
    """
    IDENTIFIER = 'horovod_cluster'
    SCHEMA = HorovodClusterSchema

    def __init__(self, master=None, worker=None):
        self.master = master
        self.worker = worker


class PytorchClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return PytorchClusterConfig


class PytorchClusterConfig(BaseConfig):
    """
    Pytorch cluster config.

    Args:
        master: list(str). The list of master host in a Pytorch cluster.
        worker: list(str). The list of worker host in a Pytorch cluster.
    """
    IDENTIFIER = 'pytorch_cluster'
    SCHEMA = PytorchClusterSchema

    def __init__(self, master=None, worker=None):
        self.master = master
        self.worker = worker


class MPIClusterSchema(BaseSchema):
    worker = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return MPIClusterConfig


class MPIClusterConfig(BaseConfig):
    """
    MPI cluster config.

    Args:
        worker: list(str). The list of worker host in a Horovod cluster.
    """
    IDENTIFIER = 'mpi_cluster'
    SCHEMA = MPIClusterSchema

    def __init__(self, worker=None):
        self.worker = worker


class MXNetClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)
    server = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return MXNetClusterConfig


class MXNetClusterConfig(BaseConfig):
    """
    MXNet cluster config.

    Args:
        master: list(str). The list of master host in a mxnet cluster.
        worker: list(str). The list of worker host in a mxnet cluster.
        server: list(str). The list of server host in a mxnet cluster.
    """
    IDENTIFIER = 'mxnet_cluster'
    SCHEMA = MXNetClusterSchema

    def __init__(self, master=None, worker=None, server=None):
        self.master = master
        self.worker = worker
        self.server = server


class FrameworkEnvironmentMixin(object):
    @staticmethod
    def _get_env_indexed_property(obj, getter):
        if not obj:
            return {}
        return {o.index: getter(o) for o in obj if getter(o)}

    # @property
    # def default_worker_config(self):
    #     return self.default_worker.config if self.default_worker else None

    @property
    def default_worker_resources(self):
        return self.default_worker.resources if self.default_worker else None

    @property
    def default_worker_node_selector(self):
        return self.default_worker.node_selector if self.default_worker else None

    @property
    def default_worker_affinity(self):
        return self.default_worker.affinity if self.default_worker else None

    @property
    def default_worker_tolerations(self):
        return self.default_worker.tolerations if self.default_worker else None

    # @property
    # def default_ps_config(self):
    #     return self.default_ps.config if self.default_ps else None

    @property
    def default_ps_resources(self):
        return self.default_ps.resources if self.default_ps else None

    @property
    def default_ps_node_selector(self):
        return self.default_ps.node_selector if self.default_ps else None

    @property
    def default_ps_affinity(self):
        return self.default_ps.affinity if self.default_ps else None

    @property
    def default_ps_tolerations(self):
        return self.default_ps.tolerations if self.default_ps else None

    # @property
    # def worker_configs(self):
    #     return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.config)

    @property
    def worker_resources(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.resources)

    @property
    def worker_node_selectors(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.node_selector)

    @property
    def worker_affinities(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.affinity)

    @property
    def worker_tolerations(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.tolerations)

    # @property
    # def ps_configs(self):
    #     return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.config)

    @property
    def ps_resources(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.resources)

    @property
    def ps_node_selectors(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.node_selector)

    @property
    def ps_affinities(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.affinity)

    @property
    def ps_tolerations(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.tolerations)


class TensorflowSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    n_ps = fields.Int(allow_none=True)
    delay_workers_by_global_step = fields.Bool(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)
    default_ps = fields.Nested(EnvironmentSchema, allow_none=True)
    worker = fields.Nested(EnvironmentSchema, many=True, allow_none=True)
    ps = fields.Nested(EnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowConfig


class TensorflowConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Tensorflow environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        n_ps: `int`. The number of ps requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        default_ps: `PodEnvironment`. The default pod environment to use for all ps.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
        ps: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific ps.
    """
    IDENTIFIER = 'tensorflow'
    SCHEMA = TensorflowSchema

    def __init__(self,
                 n_workers=0,
                 n_ps=0,
                 delay_workers_by_global_step=False,
                 run_config=None,
                 default_worker=None,
                 default_ps=None,
                 worker=None,
                 ps=None,
                 ):
        self.n_workers = n_workers
        self.n_ps = n_ps
        self.delay_workers_by_global_step = delay_workers_by_global_step
        self.run_config = run_config
        self.default_worker = default_worker
        self.default_ps = default_ps
        self.worker = worker
        self.ps = ps


class HorovodSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)
    worker = fields.Nested(EnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return HorovodConfig


class HorovodConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Horovod environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
    """
    IDENTIFIER = 'horovod'
    SCHEMA = HorovodSchema

    def __init__(self,
                 n_workers=0,
                 default_worker=None,
                 worker=None,
                 ):
        self.n_workers = n_workers
        self.default_worker = default_worker
        self.worker = worker


class MPISchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return MPIConfig


class MPIConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    MPI job environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
    """
    IDENTIFIER = 'mpi'
    SCHEMA = MPISchema

    def __init__(self,
                 n_workers=0,
                 default_worker=None,
                 ):
        self.n_workers = n_workers
        self.default_worker = default_worker
        self.worker = None


class PytorchSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)
    worker = fields.Nested(EnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return PytorchConfig


class PytorchConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Pytorch environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
    """
    IDENTIFIER = 'pytorch'
    SCHEMA = PytorchSchema

    def __init__(self,
                 n_workers=0,
                 default_worker=None,
                 worker=None,
                 ):
        self.n_workers = n_workers
        self.default_worker = default_worker
        self.worker = worker


class MXNetSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    n_ps = fields.Int(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)
    default_ps = fields.Nested(EnvironmentSchema, allow_none=True)
    worker = fields.Nested(EnvironmentSchema, many=True, allow_none=True)
    ps = fields.Nested(EnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return MXNetConfig


class MXNetConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    MXNet environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        n_ps: `int`. The number of ps requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        default_ps: `PodEnvironment`. The default pod environment to use for all ps.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
        ps: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific ps.
    """
    IDENTIFIER = 'mxnet'
    SCHEMA = MXNetSchema

    def __init__(self,
                 n_workers=0,
                 n_ps=0,
                 default_worker=None,
                 default_ps=None,
                 worker=None,
                 ps=None,
                 ):
        self.n_workers = n_workers
        self.n_ps = n_ps
        self.default_worker = default_worker
        self.default_ps = default_ps
        self.worker = worker
        self.ps = ps


class ReplicasSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    n_ps = fields.Int(allow_none=True)
    default_worker = fields.Nested(EnvironmentSchema, allow_none=True)
    default_ps = fields.Nested(EnvironmentSchema, allow_none=True)
    worker = fields.Nested(EnvironmentSchema, many=True, allow_none=True)
    ps = fields.Nested(EnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return ReplicasConfig


class ReplicasConfig(BaseConfig):
    """
    Replicas environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        n_ps: `int`. The number of ps requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        default_ps: `PodEnvironment`. The default pod environment to use for all ps.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
        ps: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific ps.
    """
    IDENTIFIER = 'replicas'
    SCHEMA = ReplicasSchema
    REDUCED_ATTRIBUTES = ['n_workers', 'n_ps', 'default_worker', 'default_ps', 'worker', 'ps']

    def __init__(self,
                 n_workers=None,
                 n_ps=None,
                 default_worker=None,
                 default_ps=None,
                 worker=None,
                 ps=None,
                 ):
        self.n_workers = n_workers
        self.n_ps = n_ps
        self.default_worker = default_worker
        self.default_ps = default_ps
        self.worker = worker
        self.ps = ps


class ExperimentEnvironmentSchema(EnvironmentSchema):
    replicas = fields.Nested(ReplicasSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentEnvironmentConfig


class ExperimentEnvironmentConfig(EnvironmentConfig):
    """
    Environment config.

    Args:
        resources: `PodResourcesConfig`. The resources config definition.
        node_selector: `dict`.
        affinity: `dict`.
        tolerations: `list(dict)`.
        replicas: `ReplicasConfig`.
    """
    IDENTIFIER = 'environment'
    SCHEMA = ExperimentEnvironmentSchema
    REDUCED_ATTRIBUTES = EnvironmentConfig.REDUCED_ATTRIBUTES + ['replicas']

    def __init__(self,
                 index=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 service_account=None,
                 image_pull_secrets=None,
                 max_restarts=None,
                 secret_refs=None,
                 config_map_refs=None,
                 configmap_refs=None,
                 data_refs=None,
                 artifact_refs=None,
                 outputs=None,
                 persistence=None,
                 replicas=None):
        super(ExperimentEnvironmentConfig, self).__init__(
            index=index,
            resources=resources,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
            service_account=service_account,
            image_pull_secrets=image_pull_secrets,
            max_restarts=max_restarts,
            secret_refs=secret_refs,
            config_map_refs=config_map_refs,
            configmap_refs=configmap_refs,
            data_refs=data_refs,
            artifact_refs=artifact_refs,
            persistence=persistence,
            outputs=outputs,
        )
        self.replicas = replicas



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.specs import kinds
from polyaxon_schemas.specs.build import BuildSpecification
from polyaxon_schemas.specs.experiment import ExperimentSpecification
from polyaxon_schemas.specs.group import GroupSpecification
from polyaxon_schemas.specs.job import JobSpecification
from polyaxon_schemas.specs.notebook import NotebookSpecification
from polyaxon_schemas.specs.pipelines import PipelineSpecification
from polyaxon_schemas.specs.tensorboard import TensorboardSpecification

SPECIFICATION_BY_KIND = {
    kinds.BUILD: BuildSpecification,
    kinds.EXPERIMENT: ExperimentSpecification,
    kinds.GROUP: GroupSpecification,
    kinds.JOB: JobSpecification,
    kinds.NOTEBOOK: NotebookSpecification,
    kinds.TENSORBOARD: TensorboardSpecification,
    kinds.PIPELINE: PipelineSpecification,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.initializations import OnesInitializerConfig, ZerosInitializerConfig
from polyaxon_schemas.ml.layers.normalization import BatchNormalizationConfig


class TestNormalizationConfigs(TestCase):
    def test_batch_normalization_config(self):
        config_dict = {
            'axis': -1,
            'momentum': 0.99,
            'epsilon': 1e-3,
            'center': True,
            'scale': True,
            'beta_initializer': ZerosInitializerConfig().to_schema(),
            'gamma_initializer': OnesInitializerConfig().to_schema(),
            'moving_mean_initializer': ZerosInitializerConfig().to_schema(),
            'moving_variance_initializer': OnesInitializerConfig().to_schema(),
            'beta_regularizer': None,
            'gamma_regularizer': None,
            'beta_constraint': None,
            'gamma_constraint': None,
        }

        config = BatchNormalizationConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validate
from marshmallow.base import FieldABC


def get_obj_or_list_obj(container, value, min_length=None, max_length=None):
    try:
        return container.deserialize(value)
    except (ValueError, TypeError, ValidationError):
        pass

    if not isinstance(value, (list, tuple)):
        raise ValidationError("This field expects an {container} or a list of {container}s.".format(
            container=container.__class__.__name__))

    value = validate.Length(min=min_length, max=max_length)(value)
    try:
        return [container.deserialize(v) for v in value]
    except (ValueError, TypeError):
        raise ValidationError("This field expects an {container} or a list of {container}s.".format(
            container=container.__class__.__name__))


class ObjectOrListObject(fields.Field):
    def __init__(self, cls_or_instance, min=None, max=None, **kwargs):  # noqa
        self.min = min
        self.max = max

        super(ObjectOrListObject, self).__init__(**kwargs)
        if isinstance(cls_or_instance, type):
            if not issubclass(cls_or_instance, FieldABC):
                raise ValueError('The type of the list elements '
                                 'must be a subclass of '
                                 'marshmallow.base.FieldABC')
            self.container = cls_or_instance()
        else:
            if not isinstance(cls_or_instance, FieldABC):
                raise ValueError('The instances of the list '
                                 'elements must be of type '
                                 'marshmallow.base.FieldABC')
            self.container = cls_or_instance

    def _deserialize(self, value, attr, data, **kwargs):
        return get_obj_or_list_obj(self.container, value, self.min, self.max)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from hestia.tz_utils import local_now
from marshmallow import ValidationError

from polyaxon_schemas.exceptions import PolyaxonSchemaError
from polyaxon_schemas.ops.io import IOTypes
from polyaxon_schemas.polyflow.ops import OpConfig
from polyaxon_schemas.polyflow.pipeline import PipelineConfig
from polyaxon_schemas.polyflow.template_ref import TemplateRefConfig


class TestPipelineConfigs(TestCase):
    def test_pipelines_base_attrs(self):
        config_dict = {'concurrency': 'foo'}
        with self.assertRaises(ValidationError):
            PipelineConfig.from_dict(config_dict)

        config_dict = {
            'concurrency': 2,
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

        config_dict = {
            'concurrency': 2,
            'execute_at': local_now().isoformat(),
            'timeout': 1000
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_wrong_pipelines_ops(self):
        config_dict = {'ops': 'foo'}
        with self.assertRaises(ValidationError):
            PipelineConfig.from_dict(config_dict)

        config_dict = {'ops': ['foo']}
        with self.assertRaises(ValidationError):
            PipelineConfig.from_dict(config_dict)

    def test_pipelines_ops(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                    'description': 'description A',
                    'tags': ['tag11', 'tag12'],
                    'params': {'param1': 'text', 'param2': 12, 'param3': 'https://foo.com'},
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 2
                },
                {
                    'template': {'url': 'https://url-to-temaplte.com'},
                    'name': 'B',
                    'description': 'description B',
                    'tags': ['tag21', 'tag22'],
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': 12,
                        'param3': 'https://foo.com'
                    },
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 2
                },
                {
                    'template': {'name': 'my-template'},
                    'name': 'C',
                    'description': 'description C',
                    'tags': ['tag31', 'tag32'],
                    'params': {
                        'param2': 12.34,
                        'param3': False
                    },
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 5,
                    'retry_delay': 12,
                    'retry_exp_backoff': True,
                },
                {
                    'template': {'path': './relative/path/to/my-template.yaml'},
                    'name': 'D',
                    'description': 'description D',
                    'tags': ['tag31', 'tag32'],
                    'dependencies': ['B', 'C'],
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 3
                    },
                    'max_retries': 3
                }
            ],
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_pipelines_templates(self):
        config_dict = {
            'templates': [
                {
                    'kind': 'experiment',
                    'name': 'experiment-template',
                    'description': 'description experiment',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'build': {'dockerfile': 'Dockerfile'},
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'group',
                    'name': 'group-template',
                    'description': 'description group',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'build': {'ref': '12'},
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'ref': '12'
                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'description': 'description job',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH,
                            'is_optional': True,
                            'default': 's3://foo'
                        }
                    ],
                    'build': {'ref': '12'},
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
            ],
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_pipelines_dag(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                    'params': {'param1': 'text', 'param2': 12, 'param3': 'https://foo.com'},
                },
                {
                    'template': {'url': 'https://url-to-temaplte.com'},
                    'name': 'B',
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': 12,
                        'param3': 'https://foo.com'
                    },
                },
                {
                    'template': {'name': 'my-template'},
                    'name': 'C',
                    'params': {
                        'param2': 12.34,
                        'param3': False
                    },
                },
                {
                    'template': {'path': './relative/path/to/my-template.yaml'},
                    'name': 'D',
                    'dependencies': ['B', 'C'],
                }
            ],
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', 'C', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] in [['A', 'C'], ['C', 'A']]
        assert sorted_dag[1] == ['B']
        assert sorted_dag[2] == ['D']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', }

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'url': 'https://url-to-temaplte.com'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'my-template'}
        assert config.dag['C'].upstream == set()
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {
            'path': './relative/path/to/my-template.yaml'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

    def test_pipelines_parallel_ops(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                },
                {
                    'template': {'name': 'bar'},
                    'name': 'D',
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', 'B', 'C', 'D'}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert len(config.sort_topologically(dag=dag)) == 1  # order can be any
        assert len(config.sort_topologically(dag=dag)[0]) == 4  # order can be any

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == set()

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == set()
        assert config.dag['B'].downstream == set()

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == set()
        assert config.dag['C'].downstream == set()

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'name': 'bar'}
        assert config.dag['D'].upstream == set()
        assert config.dag['D'].downstream == set()

    def test_pipelines_sequential_ops(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['B']
                },
                {
                    'template': {'name': 'bar'},
                    'name': 'D',
                    'dependencies': ['C']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [['A'], ['B'], ['C'], ['D']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', }

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'C', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'B', }
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'name': 'bar'}
        assert config.dag['D'].upstream == {'C', }
        assert config.dag['D'].downstream == set()

    def test_pipelines_all_downstream(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                    'params': {'param1': 'text', 'param2': 12, 'param3': 'https://foo.com'},
                    'dependencies': ['B', 'C', 'D'],
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                },
                {
                    'template': {'event': 'event2'},
                    'name': 'C',
                },
                {
                    'template': {'event': 'event3'},
                    'name': 'D',
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'B', 'C', 'D', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        # Sort is not guaranteed at stage 0
        assert len(sorted_dag[0]) == 3
        assert set(sorted_dag[0]) == {'B', 'C', 'D'}
        assert sorted_dag[1] == ['A']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == {'B', 'C', 'D'}
        assert config.dag['A'].downstream == set()

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == set()
        assert config.dag['B'].downstream == {'A', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'event': 'event2'}
        assert config.dag['C'].upstream == set()
        assert config.dag['C'].downstream == {'A', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'event': 'event3'}
        assert config.dag['D'].upstream == set()
        assert config.dag['D'].downstream == {'A', }

    def test_pipelines_acyclic_deps(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                    'dependencies': ['B']
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['A']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == set()
        assert config.get_orphan_ops(dag=dag) == set()
        with self.assertRaises(PolyaxonSchemaError):
            config.sort_topologically(dag=dag)

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == {'B', }
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'A', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == set()

    def test_pipelines_circular_deps(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                    'dependencies': ['C']
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['B']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == set()
        assert config.get_orphan_ops(dag=dag) == set()
        with self.assertRaises(PolyaxonSchemaError):
            config.sort_topologically(dag=dag)

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == {'C', }
        assert config.dag['A'].downstream == {'B', }

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'C', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'B', }
        assert config.dag['C'].downstream == {'A', }

    def test_pipelines_adding_ops_one_by_one_manually(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['A']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']  # Only this one has consistent order
        assert sorted_dag[1] in [['B', 'C'], ['C', 'B']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == set()

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == set()

        template4 = TemplateRefConfig.from_dict({'action': 'action4'})
        operationD = OpConfig(template=template4, name='D', dependencies=['B', 'C'])
        config.add_op(operationD)

        operationE = OpConfig(template=template4, name='E', dependencies=['A'])
        config.add_op(operationE)

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {'B', 'C', 'E'}
        assert sorted_dag[2] == ['D']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C', 'E'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

        assert config.dag['E'].op.name == 'E'
        assert config.dag['E'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['E'].upstream == {'A', }
        assert config.dag['E'].downstream == set()

        # Adding the same ops should not alter the behaviour
        config.add_ops([operationD, operationE])

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {'B', 'C', 'E'}
        assert sorted_dag[2] == ['D']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C', 'E'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

        assert config.dag['E'].op.name == 'E'
        assert config.dag['E'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['E'].upstream == {'A', }
        assert config.dag['E'].downstream == set()

    def test_pipelines_adding_ops_many_manually(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['A']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']  # Only this one has consistent order
        assert sorted_dag[1] in [['B', 'C'], ['C', 'B']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == set()

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == set()

        template4 = TemplateRefConfig.from_dict({'action': 'action4'})
        operationD = OpConfig(template=template4, name='D', dependencies=['B', 'C'])
        operationE = OpConfig(template=template4, name='E', dependencies=['A'])
        config.add_ops([operationD, operationE])

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        assert len(config.dag) == 5
        assert config.get_independent_ops(dag=config.dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {'B', 'C', 'E'}
        assert sorted_dag[2] == ['D']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C', 'E'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

        assert config.dag['E'].op.name == 'E'
        assert config.dag['E'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['E'].upstream == {'A', }
        assert config.dag['E'].downstream == set()

        # Adding the same ops should not alter the behaviour
        config.add_op(operationD)
        config.add_op(operationE)

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {'B', 'C', 'E'}
        assert sorted_dag[2] == ['D']

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C', 'E'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', }
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

        assert config.dag['E'].op.name == 'E'
        assert config.dag['E'].op.template.to_dict() == {'action': 'action4'}
        assert config.dag['E'].upstream == {'A', }
        assert config.dag['E'].downstream == set()

    def test_pipelines_dependency_from_params(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'echo'},
                    'name': 'A',
                    'params': {'param1': 'text', 'param2': 12, 'param3': 'https://foo.com'},
                },
                {
                    'template': {'action': 'echo'},
                    'name': 'B',
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': 12,
                        'param3': 'https://foo.com'
                    },
                },
                {
                    'template': {'action': 'echo'},
                    'name': 'C',
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': '{{ ops.B.outputs.x }}',
                    },

                },
                {
                    'template': {'action': 'echo'},
                    'name': 'D',
                    'dependencies': ['B', 'C'],
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [['A'], ['B'], ['C'], ['D']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'C', 'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['C'].upstream == {'A', 'B'}
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

    def test_pipelines_dependency_and_params(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'echo'},
                    'name': 'A',
                    'params': {'param1': 'text', 'param2': 12, 'param3': 'https://foo.com'},
                },
                {
                    'template': {'action': 'echo'},
                    'name': 'B',
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': 12,
                        'param3': 'https://foo.com'
                    },
                    'dependencies': ['A']
                },
                {
                    'template': {'action': 'echo'},
                    'name': 'C',
                    'params': {
                        'param2': '{{ ops.B.outputs.x }}',
                    },
                    'dependencies': ['A']
                },
                {
                    'template': {'action': 'echo'},
                    'name': 'D',
                    'dependencies': ['B', 'C'],
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [['A'], ['B'], ['C'], ['D']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == {'C', 'D', }

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['C'].upstream == {'A', 'B'}
        assert config.dag['C'].downstream == {'D', }

        assert config.dag['D'].op.name == 'D'
        assert config.dag['D'].op.template.to_dict() == {'action': 'echo'}
        assert config.dag['D'].upstream == {'B', 'C'}
        assert config.dag['D'].downstream == set()

    def test_pipelines_orphan_ops(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'event': 'event1'},
                    'name': 'B',
                    'dependencies': ['A']
                },
                {
                    'template': {'name': 'foo'},
                    'name': 'C',
                    'dependencies': ['A', 'E']
                }
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {'A', 'E', }
        assert config.get_orphan_ops(dag=dag) == {'E', }
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] in [['A', 'E'], ['E', 'A']]
        assert sorted_dag[1] in [['B', 'C'], ['C', 'B']]

        assert config.dag['A'].op.name == 'A'
        assert config.dag['A'].op.template.to_dict() == {'action': 'action1'}
        assert config.dag['A'].upstream == set()
        assert config.dag['A'].downstream == {'B', 'C'}

        assert config.dag['B'].op.name == 'B'
        assert config.dag['B'].op.template.to_dict() == {'event': 'event1'}
        assert config.dag['B'].upstream == {'A', }
        assert config.dag['B'].downstream == set()

        assert config.dag['C'].op.name == 'C'
        assert config.dag['C'].op.template.to_dict() == {'name': 'foo'}
        assert config.dag['C'].upstream == {'A', 'E'}
        assert config.dag['C'].downstream == set()

    def test_pipelines_with_duplicate_template_names(self):
        config_dict = {
            'ops': [
                {
                    'template': {'action': 'action1'},
                    'name': 'A',
                },
                {
                    'template': {'action': 'action1'},
                    'name': 'B',
                }
            ],
            'templates': [
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'ref': '12'
                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_pipelines_with_duplicate_op_names(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template1'},
                    'name': 'A',
                },
                {
                    'template': {'name': 'build-template1'},
                    'name': 'A',
                }
            ],
            'templates': [
                {
                    'kind': 'build',
                    'name': 'build-template1',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'ref': '12'
                },
                {
                    'kind': 'build',
                    'name': 'build-template2',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_pipelines_with_op_requesting_undefined_template(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template1'},
                    'name': 'A',
                },
                {
                    'template': {'name': 'build-template1'},
                    'name': 'A',
                }
            ],
            'templates': [
                {
                    'kind': 'build',
                    'name': 'build-template2',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(PolyaxonSchemaError):
            config.process_templates()

    def test_pipelines_with_template_not_defining_inputs(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template'},
                    'name': 'A',
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A']
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',

                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_templates()

    def test_pipelines_with_template_not_defining_inputs_and_ops_with_params(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template'},
                    'name': 'A',
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                        'param2': 12,
                        'param3': 'https://foo.com'
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',

                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_ops_template_required_inputs(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                },
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                    ],
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_ops_template_optional_inputs(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                },
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                            'is_optional': True,
                            'default': 12.2
                        },
                    ],
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_templates()

    def test_pipelines_with_ops_template_optional_inputs_and_wrong_param(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                    'params': {
                        'input1': 'foo',
                    },
                },
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                            'is_optional': True,
                            'default': 12.2
                        },
                    ],
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_ops_template_validation(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                    'params': {
                        'input1': 'sdf',
                        'input2': 12.,
                        'input3': False
                    },
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'input1': 'ooo',
                        'input2': 12.123,
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.STR,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                        {
                            'name': 'input3',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_templates()

    def test_pipelines_with_wrong_refs(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                    'params': {
                        'input1': 'sdf',
                        'input2': 12.,
                        'input3': False
                    },
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'input1': '{{ ops.A.outputs.output1 }}',
                        'input2': 12.123,
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.STR,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                        {
                            'name': 'input3',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_correct_refs(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                    'params': {
                        'input1': 2,
                        'input2': 'gs://bucket/path/to/blob/',
                    },
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'input1': '{{ ops.A.outputs.output1 }}',
                        'input2': 'gs://bucket/path/to/blob/',
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.INT,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.GCS_PATH,
                        },
                        {
                            'name': 'input3',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.INT,
                            'is_optional': True,
                            'default': 123
                        },
                    ]
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_templates()

    def test_pipelines_with_correct_ref_and_wrong_ref_type(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'job-template'},
                    'name': 'A',
                    'params': {
                        'input1': 2,
                        'input2': 'gs://bucket/path/to/blob/',
                        'output1': 123
                    },
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'input1': 3,
                        'input2': '{{ ops.A.outputs.output1 }}',
                        'output1': 123
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.INT,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.GCS_PATH,
                        },
                        {
                            'name': 'input3',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.INT,
                        },
                    ]
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_template_not_defining_inputs_and_ops_refs_params(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template'},
                    'name': 'A',
                },
                {
                    'template': {'name': 'job-template'},
                    'name': 'B',
                    'dependencies': ['A'],
                    'params': {
                        'param1': '{{ ops.A.outputs.x }}',
                    },
                }
            ],
            'templates': [
                {
                    'kind': 'job',
                    'name': 'job-template',

                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                },
            ]
        }
        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_templates()

    def test_pipelines_with_ops_and_templates(self):
        config_dict = {
            'ops': [
                {
                    'template': {'name': 'build-template'},
                    'name': 'A',
                    'description': 'description A',
                    'tags': ['tag11', 'tag12'],
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 2
                },
                {
                    'template': {'name': 'experiment-template'},
                    'name': 'B',
                    'description': 'description B',
                    'tags': ['tag21', 'tag22'],
                    'dependencies': ['A'],
                    'params': {
                        'input1': 11.1,
                        'input2': False,
                        'output1': 'S3://foo.com'
                    },
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 2
                },
                {
                    'template': {'name': 'group-template'},
                    'name': 'C',
                    'description': 'description C',
                    'tags': ['tag31', 'tag32'],
                    'params': {
                        'input1': '{{ ops.B.outputs.output1 }}',
                        'output1': 'S3://foo.com',
                    },
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                    },
                    'max_retries': 5,
                    'retry_delay': 12,
                    'retry_exp_backoff': True,
                },
            ],
            'templates': [
                {
                    'kind': 'experiment',
                    'name': 'experiment-template',
                    'description': 'description experiment',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.FLOAT,
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'build': {'dockerfile': 'Dockerfile'},
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'group',
                    'name': 'group-template',
                    'description': 'description group',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        },
                        {
                            'name': 'input2',
                            'description': 'some text',
                            'type': IOTypes.BOOL,
                            'is_optional': True,
                            'default': True,
                        }
                    ],
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'build': {'ref': '12'},
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'build',
                    'name': 'build-template',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'ref': '12'
                },
                {
                    'kind': 'build',
                    'name': 'build-template2',
                    'description': 'description build',
                    'tags': ['tag11', 'tag12'],
                    'backend': 'kaniko',
                    'dockerfile': 'Dockerfile',
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
                {
                    'kind': 'job',
                    'name': 'job-template',
                    'description': 'description job',
                    'tags': ['tag11', 'tag12'],
                    'inputs': [
                        {
                            'name': 'input1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH,
                            'is_optional': True,
                            'default': 's3://foo'
                        }
                    ],
                    'build': {'ref': '12'},
                    'outputs': [
                        {
                            'name': 'output1',
                            'description': 'some text',
                            'type': IOTypes.S3_PATH
                        }
                    ],
                    'environment': {
                        'resources': {'cpu': {'requests': 1}},
                        'node_selector': {'polyaxon': 'core'},
                        'service_account': 'service',
                        'image_pull_secrets': ['secret1', 'secret2'],
                        'max_restarts': 2
                    },
                },
            ],
        }

        config = PipelineConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        config.process_templates()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {'A', }
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] == ['A']
        assert sorted_dag[1] == ['B']
        assert sorted_dag[2] == ['C']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.initializations import UniformInitializerConfig
from polyaxon_schemas.ml.layers.embeddings import EmbeddingConfig


class TestEmbeddingConfigs(TestCase):
    def test_embedding_config(self):
        config_dict = {
            'input_dim': 100,
            'output_dim': 100,
            'embeddings_initializer': UniformInitializerConfig().to_schema(),
            'embeddings_regularizer': None,
            'activity_regularizer': None,
            'embeddings_constraint': None,
            'mask_zero': False,
            'input_length': None,
        }
        config = EmbeddingConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ml.regularizations import (
    L1L2RegularizerConfig,
    L1RegularizerConfig,
    L2RegularizerConfig
)


class TestRegularizationConfigs(TestCase):
    def test_l1_config(self):
        config_dict = {
            'l': 0.5,
            'name': 'l1',
            'collect': False
        }
        config = L1RegularizerConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_l2_config(self):
        config_dict = {
            'l': 0.5,
            'name': 'l2',
            'collect': False
        }
        config = L2RegularizerConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_l2l1_config(self):
        config_dict = {
            'l1': 0.5,
            'l2': 0.5,
            'name': 'l2l1',
            'collect': True
        }
        config = L1L2RegularizerConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import ast
import copy
import jinja2
import numpy as np
import six

from collections import Mapping, defaultdict

from hestia.list_utils import to_list
from rhea.utils import deep_update

from polyaxon_schemas.exceptions import PolyaxonfileError


class Parser(object):
    """Parses the Polyaxonfile."""

    env = jinja2.Environment()

    @classmethod
    def get_headers(cls, spec, data):
        parsed_data = {
            section: data[section] for section in spec.HEADER_SECTIONS
            if data.get(section)
        }
        return parsed_data

    @staticmethod
    def _get_section(config, section):
        if not hasattr(config, section):
            return None
        section_data = getattr(config, section)
        if hasattr(section_data, 'to_dict'):
            return section_data.to_dict()
        return section_data

    @classmethod
    def parse(cls, spec, config, params, matrix_params=None):  # pylint:disable=too-many-branches
        params = params or {}
        matrix_params = copy.copy(matrix_params)
        if matrix_params:
            params = deep_update(matrix_params, params)

        parsed_data = {
            spec.VERSION: config.version,
            spec.KIND: config.kind,
        }

        if params:
            params = cls.parse_expression(spec, params, params)
            parsed_data[spec.PARAMS] = params

        for section in spec.STD_PARSING_SECTIONS:
            config_section = cls._get_section(config, section)
            if config_section:
                parsed_data[section] = cls.parse_expression(spec, config_section, params)

        for section in spec.OP_PARSING_SECTIONS:
            config_section = cls._get_section(config, section)
            if config_section:
                parsed_data[section] = cls.parse_expression(
                    spec, config_section, params, True, False)

        config_section = cls._get_section(config, spec.RUN)
        if config_section:
            parsed_data[spec.RUN] = cls.parse_expression(spec, config_section, params, True, False)

        for section in spec.GRAPH_SECTIONS:
            config_section = cls._get_section(config, section)
            if config_section:
                parsed_data[section] = cls.parse_expression(
                    spec, config_section, params, True, True)

        return parsed_data

    @classmethod
    def parse_expression(cls,  # pylint:disable=too-many-branches
                         spec,
                         expression,
                         params,
                         check_operators=False,
                         check_graph=False):
        if isinstance(expression, (int, float, complex, type(None))):
            return expression
        if isinstance(expression, np.integer):
            return int(expression)
        if isinstance(expression, np.floating):
            return float(expression)
        if isinstance(expression, Mapping):
            if len(expression) == 1:
                old_key, value = list(six.iteritems(expression))[0]
                # always parse the keys, they must be base object or evaluate to base objects
                key = cls.parse_expression(spec, old_key, params)
                if check_operators and cls.is_operator(spec, key):
                    return cls._parse_operator(spec, {key: value}, params)
                if check_graph and key in ['graph', 'encoder', 'decoder']:
                    return {key: cls._parse_graph(spec, value, params)}
                if check_graph and key == 'feature_processors':  # noqa, no-else-return
                    return {
                        key: {
                            cls.parse_expression(spec, f_key, params):
                                cls._parse_graph(spec, f_vlaue, params)
                            for f_key, f_vlaue in six.iteritems(value)
                        }
                    }
                else:
                    return {
                        key: cls.parse_expression(
                            spec, value, params, check_operators, check_graph)
                    }

            new_expression = {}
            for k, v in six.iteritems(expression):
                new_expression.update(
                    cls.parse_expression(spec, {k: v}, params, check_operators, check_graph))
            return new_expression

        if isinstance(expression, list):
            return list(cls.parse_expression(spec, v, params, check_operators, check_graph)
                        for v in expression)
        if isinstance(expression, tuple):
            return tuple(cls.parse_expression(spec, v, params, check_operators, check_graph)
                         for v in expression)
        if isinstance(expression, six.string_types):
            return cls._evaluate_expression(
                spec, expression, params, check_operators, check_graph)

    @classmethod
    def _evaluate_expression(cls, spec, expression, params, check_operators, check_graph):
        result = cls.env.from_string(expression).render(**params)
        if result == expression:
            try:
                return ast.literal_eval(result)
            except (ValueError, SyntaxError):
                pass
            return result
        return cls.parse_expression(spec, result, params, check_operators, check_graph)

    @classmethod
    def _parse_operator(cls, spec, expression, params):
        k, v = list(six.iteritems(expression))[0]
        op = spec.OPERATORS[k].from_dict(v)
        return op.parse(spec=spec, parser=cls, params=params)

    @staticmethod
    def is_operator(spec, key):
        return key in spec.OPERATORS

    @classmethod
    def _parse_graph(cls, spec, graph, params):  # noqa, too-many-branches
        input_layers = to_list(graph['input_layers'])
        layer_names = set(input_layers)
        tags = {}
        layers = []
        outputs = []
        layers_counters = defaultdict(int)
        unused_layers = set(input_layers)

        if not isinstance(graph['layers'], list):
            raise PolyaxonfileError("Graph definition expects a list of layer definitions.")

        def add_tag(tag, layer_value):
            if tag in tags:
                tags[tag] = to_list(tags[tag])
                tags[tag].append(layer_value['name'])
            else:
                tags[tag] = layer_value['name']

        def get_layer_name(layer_value, layer_type):
            if 'name' not in layer_value:
                layers_counters[layer_type] += 1
                return '{}_{}'.format(layer_type, layers_counters[layer_type])

            return layer_value['name']

        layers_params = {}
        layers_params.update(params)

        last_layer = None
        first_layer = True
        for layer_expression in graph['layers']:
            parsed_layer = cls.parse_expression(spec, layer_expression, layers_params, True)
            # Gather all tags from the layers
            parsed_layer = to_list(parsed_layer)
            for layer in parsed_layer:
                if not layer:
                    continue

                layer_type, layer_value = list(six.iteritems(layer))[0]

                if layer_value is None:
                    layer_value = {}
                # Check that the layer has a name otherwise generate one
                name = get_layer_name(layer_value, layer_type)
                if name not in layer_names:
                    layer_names.add(name)
                    layer_value['name'] = name
                else:
                    raise PolyaxonfileError(
                        "The name `{}` is used 2 times in the graph. "
                        "All layer names should be unique. "
                        "If you need to reference a layer in a for loop "
                        "think about using `tags`".format(name))

                for tag in to_list(layer_value.get('tags', [])):
                    add_tag(tag, layer_value)

                # Check if the layer is an output
                if layer_value.get('is_output', False) is True:
                    outputs.append(layer_value['name'])
                else:
                    # Add the layer to unused
                    unused_layers.add(layer_value['name'])

                # Check the layers inputs
                if not layer_value.get('inbound_nodes'):
                    if last_layer is not None:
                        layer_value['inbound_nodes'] = [
                            last_layer['name']]  # noqa, unsubscriptable-object
                    if first_layer and len(input_layers) == 1:
                        layer_value['inbound_nodes'] = input_layers
                    if first_layer and len(input_layers) > 1:
                        raise PolyaxonfileError("The first layer must indicate which input to use,"
                                                "You have {} layers: {}".format(len(input_layers),
                                                                                input_layers))

                first_layer = False
                for input_layer in layer_value.get('inbound_nodes', []):
                    if input_layer not in layer_names:
                        raise PolyaxonfileError(
                            "The layer `{}` has a non existing "
                            "inbound node `{}`".format(layer_value['name'], input_layer))
                    if input_layer in unused_layers:
                        unused_layers.remove(input_layer)

                # Add layer
                layers.append({layer_type: layer_value})

                # Update layers_params
                layers_params['tags'] = tags

                # Update last_layer
                last_layer = layer_value

        # Add last layer as output
        if last_layer:
            if last_layer['name'] not in outputs:
                outputs.append(last_layer['name'])

            # Remove last layer from unused layers
            if last_layer['name'] in unused_layers:
                unused_layers.remove(last_layer['name'])

        # Check if some layers are unused
        if unused_layers:
            raise PolyaxonfileError(
                "These layers `{}` were declared but are not used.".format(unused_layers))

        return {
            'input_layers': to_list(graph['input_layers']),
            'layers': layers,
            'output_layers': outputs
        }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ops.run import BaseRunConfig, BaseRunSchema


class TensorboardSchema(BaseRunSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('tensorboard'))

    @staticmethod
    def schema_config():
        return TensorboardConfig


class TensorboardConfig(BaseRunConfig):
    IDENTIFIER = 'tensorboard'
    SCHEMA = TensorboardSchema



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

from unittest import TestCase

from flaky import flaky

from polyaxon_schemas.exceptions import PolyaxonfileError
from polyaxon_schemas.ops.build_job import BuildConfig
from polyaxon_schemas.ops.environments.pods import EnvironmentConfig
from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig
from polyaxon_schemas.ops.experiment.backends import ExperimentBackend
from polyaxon_schemas.ops.experiment.frameworks import ExperimentFramework
from polyaxon_schemas.ops.group.early_stopping_policies import EarlyStoppingConfig
from polyaxon_schemas.ops.group.hptuning import HPTuningConfig, SearchAlgorithms
from polyaxon_schemas.ops.group.matrix import MatrixConfig
from polyaxon_schemas.ops.logging import LoggingConfig
from polyaxon_schemas.ops.run import RunConfig
from polyaxon_schemas.polyaxonfile import PolyaxonFile
from polyaxon_schemas.specs.frameworks import (
    HorovodSpecification,
    MPISpecification,
    MXNetSpecification,
    PytorchSpecification,
    TensorflowSpecification
)
from polyaxon_schemas.utils import TaskType


class TestPolyaxonfileWithTypes(TestCase):
    def test_using_untyped_params_raises(self):
        with self.assertRaises(PolyaxonfileError):
            PolyaxonFile(os.path.abspath('tests/fixtures/typing/untyped_params.yml'))

    def test_no_params_for_required_inputs_outputs_raises(self):
        with self.assertRaises(PolyaxonfileError):
            PolyaxonFile(os.path.abspath(
                'tests/fixtures/typing/required_inputs.yml'))

        PolyaxonFile(os.path.abspath(
            'tests/fixtures/typing/required_outputs.yml'))

    def test_required_inputs_with_params(self):
        plxfile = PolyaxonFile(os.path.abspath('tests/fixtures/typing/required_inputs.yml'),
                               params={'loss': 'bar', 'flag': False})
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.logging is None
        assert set(spec.tags) == {'foo', 'bar'}
        assert spec.params == {'loss': 'bar', 'flag': ''}
        assert spec.build.image == 'my_image'
        assert spec.run.cmd == 'video_prediction_train --loss=bar '
        assert spec.environment is None
        assert spec.framework is None
        assert spec.is_experiment
        assert spec.cluster_def == ({TaskType.MASTER: 1}, False)
        assert spec.is_experiment is True

        plxfile = PolyaxonFile(os.path.abspath('tests/fixtures/typing/required_inputs.yml'),
                               params={'loss': 'bar', 'flag': True})
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.logging is None
        assert set(spec.tags) == {'foo', 'bar'}
        assert spec.params == {'loss': 'bar', 'flag': '--flag'}
        assert spec.build.image == 'my_image'
        assert spec.run.cmd == 'video_prediction_train --loss=bar --flag'
        assert spec.environment is None
        assert spec.framework is None
        assert spec.is_experiment
        assert spec.cluster_def == ({TaskType.MASTER: 1}, False)
        assert spec.is_experiment is True

        # Adding extra value raises
        with self.assertRaises(PolyaxonfileError):
            PolyaxonFile(os.path.abspath('tests/fixtures/typing/required_inputs.yml'),
                         params={'loss': 'bar', 'value': 1.1})

        # Adding non valid params raises
        with self.assertRaises(PolyaxonfileError):
            PolyaxonFile(os.path.abspath('tests/fixtures/typing/required_inputs.yml'),
                         params={'value': 1.1})

    def test_matrix_file_passes_int_float_types(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/typing/matrix_file_with_int_float_types.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.is_group
        assert isinstance(spec.hptuning.matrix['param1'], MatrixConfig)
        assert isinstance(spec.hptuning.matrix['param2'], MatrixConfig)
        assert spec.hptuning.matrix['param1'].to_dict() == {'values': [1, 2]}
        assert spec.hptuning.matrix['param2'].to_dict() == {'values': [3.3, 4.4]}
        assert spec.matrix_space == 4
        assert isinstance(spec.hptuning, HPTuningConfig)
        assert spec.hptuning.concurrency == 2
        assert spec.search_algorithm == SearchAlgorithms.GRID
        assert spec.hptuning.early_stopping is None
        assert spec.early_stopping == []

        assert spec.experiments_def == {
            'search_algorithm': SearchAlgorithms.GRID,
            'early_stopping': False,
            'concurrency': 2,
        }

        build = spec.build
        assert build is None

        spec = spec.get_experiment_spec(matrix_declaration=spec.matrix_declaration_test)
        spec.parse_data()
        assert spec.environment is None
        assert spec.framework is None
        assert spec.cluster_def == ({TaskType.MASTER: 1}, False)
        assert spec.run.cmd == 'train --param1={param1} --param2={param2} --param3=23423'.format(
            **spec.params
        )

    def test_run_simple_file_passes(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/typing/run_cmd_simple_file.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.logging is None
        assert sorted(spec.tags) == sorted(['foo', 'bar'])
        assert spec.is_experiment
        assert isinstance(spec.build, BuildConfig)
        assert isinstance(spec.run, RunConfig)
        assert spec.environment is None
        assert spec.framework is None
        assert spec.cluster_def == ({TaskType.MASTER: 1}, False)
        run = spec.run
        assert isinstance(run, RunConfig)
        assert run.cmd == "video_prediction_train --num_masks=2 --loss=MeanSquaredError"

    def test_run_with_refs(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/typing/run_with_refs.yml'))
        spec = plxfile.specification
        required_refs = spec.raw_config.get_params_with_refs()
        assert len(required_refs) == 1
        assert required_refs[0].name == 'model_path'
        assert required_refs[0].value == 'jobs.1.outputs.doo'
        spec.parse_data(context={'jobs__1__outputs__doo': 'model_path'})
        assert spec.version == 1
        assert spec.logging is None
        assert sorted(spec.tags) == sorted(['foo', 'bar'])
        assert spec.is_experiment
        assert isinstance(spec.build, BuildConfig)
        assert isinstance(spec.run, RunConfig)
        assert spec.environment is None
        assert spec.framework is None
        assert spec.cluster_def == ({TaskType.MASTER: 1}, False)
        run = spec.run
        assert isinstance(run, RunConfig)
        assert run.cmd == "video_prediction_train --num_masks=2 --model_path=model_path"

    def test_jupyter_lab_job_with_node_selectors(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/typing/jupyterlab_with_custom_environment.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.is_notebook
        assert spec.is_notebook is True
        assert spec.backend == 'lab'
        assert spec.logging is None
        assert sorted(spec.tags) == sorted(['foo', 'bar'])
        assert isinstance(spec.build, BuildConfig)
        assert isinstance(spec.environment, EnvironmentConfig)
        assert spec.artifact_refs == ['outputs1']
        assert spec.data_refs == ['data1', 'data2']
        assert spec.secret_refs == ['secret1', 'secret2']
        assert spec.config_map_refs == ['config_map1', 'config_map2']

        node_selector = {'polyaxon.com': 'node_for_notebook_jobs'}
        assert spec.environment.node_selector == node_selector
        assert spec.node_selector == node_selector

        resources = {
            'cpu': {'requests': 1, 'limits': 2},
            'memory': {'requests': 200, 'limits': 200},
        }
        assert spec.environment.resources.to_dict() == resources
        assert spec.resources.to_dict() == resources

        affinity = {
            'nodeAffinity': {'requiredDuringSchedulingIgnoredDuringExecution': {}}
        }
        assert spec.environment.affinity == affinity
        assert spec.affinity == affinity

        tolerations = [{'key': 'key', 'operator': 'Exists'}]

        assert spec.environment.tolerations == tolerations
        assert spec.tolerations == tolerations



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validate, validates_schema

from polyaxon_schemas.base import BaseConfig
from polyaxon_schemas.ops.experiment.backends import ExperimentBackend
from polyaxon_schemas.ops.experiment.environment import (
    ExperimentEnvironmentSchema,
    HorovodConfig,
    MPIConfig,
    MXNetConfig,
    PytorchConfig,
    TensorflowConfig
)
from polyaxon_schemas.ops.experiment.frameworks import ExperimentFramework
from polyaxon_schemas.ops.run import BaseRunConfig, BaseRunSchema, RunSchema


def validate_backend(backend):
    if backend and backend not in ExperimentBackend.VALUES:
        raise ValidationError('Experiment backend `{}` not supported'.format(backend))


def validate_replicas(framework, replicas):
    if replicas and framework and framework not in ExperimentFramework.VALUES:
        raise ValidationError(
            'Distributed Experiment framework `{}` not supported'.format(framework))

    if replicas and not framework:
        raise ValidationError(
            'You must specify which framework to use for distributed experiments.')

    config = replicas.to_light_dict() if isinstance(replicas, BaseConfig) else replicas

    if framework == 'tensorflow':
        TensorflowConfig.from_dict(config)
    if framework == 'horovod':
        HorovodConfig.from_dict(config)
    if framework == 'mpi':
        MPIConfig.from_dict(config)
    if framework == 'mxnet':
        MXNetConfig.from_dict(config)
    if framework == 'pytorch':
        PytorchConfig.from_dict(config)


class ExperimentSchema(BaseRunSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('experiment'))
    environment = fields.Nested(ExperimentEnvironmentSchema, allow_none=True)
    backend = fields.Str(allow_none=True, validate=validate.OneOf(ExperimentBackend.VALUES))
    framework = fields.Str(allow_none=True)
    run = fields.Nested(RunSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentConfig

    @validates_schema
    def validate_backend(self, data):
        """Validate backend"""
        validate_backend(data.get('backend'))

    @validates_schema
    def validate_replicas(self, data):
        """Validate distributed experiment"""
        environment = data.get('environment')
        if environment and environment.replicas:
            validate_replicas(data.get('framework'), environment.replicas)


class ExperimentConfig(BaseRunConfig):
    SCHEMA = ExperimentSchema
    IDENTIFIER = 'experiment'
    REDUCED_ATTRIBUTES = BaseRunConfig.REDUCED_ATTRIBUTES + [
        'backend',
        'framework',
        'run',
    ]

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None,
                 build=None,
                 backend=None,
                 framework=None,
                 run=None,
                 ):
        validate_backend(backend)
        self.replicas = None
        if framework and environment and environment.replicas:
            validate_replicas(framework, environment.replicas)
            self.replicas = environment.replicas

        super(ExperimentConfig, self).__init__(
            version=version,
            kind=kind,
            logging=logging,
            name=name,
            description=description,
            tags=tags,
            environment=environment,
            params=params,
            declarations=declarations,
            inputs=inputs,
            outputs=outputs,
            build=build
        )
        self.backend = backend
        self.framework = framework
        self.run = run
        self.tensorflow = self.get_tensorflow()
        self.horovod = self.get_horovod()
        self.mxnet = self.get_mxnet()
        self.pytorch = self.get_pytorch()
        self.mpi = self.get_mpi()

    def has_framework(self, framework):
        return (
            self.framework == framework and
            self.replicas and
            self.backend != ExperimentBackend.MPI
        )

    def get_tensorflow(self):
        if self.has_framework(framework=ExperimentFramework.TENSORFLOW):
            return TensorflowConfig.from_dict(self.replicas.to_light_dict())

    def get_horovod(self):
        if self.has_framework(framework=ExperimentFramework.HOROVOD):
            return HorovodConfig.from_dict(self.replicas.to_light_dict())

    def get_mxnet(self):
        if self.has_framework(framework=ExperimentFramework.MXNET):
            return MXNetConfig.from_dict(self.replicas.to_light_dict())

    def get_pytorch(self):
        if self.has_framework(framework=ExperimentFramework.PYTORCH):
            return PytorchConfig.from_dict(self.replicas.to_light_dict())

    def get_mpi(self):
        if self.backend == ExperimentBackend.MPI and self.replicas:
            return MPIConfig.from_dict(self.replicas.to_light_dict())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema


class PersistenceSchema(BaseSchema):
    data = fields.List(fields.Str(), allow_none=True)
    outputs = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return PersistenceConfig


class PersistenceConfig(BaseConfig):
    """
    Persistence config.

    Defines the list of persistent volumes to mount for this specific run.

    Args:
        data: `list(str)`. The list of the names of data persistence to mount.
        outputs: `list(str)`. The list of the names of outputs persistence to mount.
    """
    IDENTIFIER = 'persistence'
    SCHEMA = PersistenceSchema

    def __init__(self, data=None, outputs=None):
        self.data = data
        self.outputs = outputs



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.fields import FloatOrStr, IntOrStr


class K8SResourcesEntrySchema(BaseSchema):
    cpu = FloatOrStr(allow_none=True)
    memory = IntOrStr(allow_none=True)
    gpu = fields.Int(allow_none=True)
    tpu = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return K8SResourcesEntryConfig


class K8SResourcesEntryConfig(BaseConfig):
    SCHEMA = K8SResourcesEntrySchema
    REDUCED_ATTRIBUTES = ['cpu', 'memory', 'gpu', 'tpu']

    def __init__(self, cpu=None, memory=None, gpu=None, tpu=None):
        self.cpu = cpu
        self.memory = memory
        self.gpu = gpu
        self.tpu = tpu


class K8SContainerResourcesSchema(BaseSchema):
    limits = fields.Nested(K8SResourcesEntrySchema, allow_none=True)
    requests = fields.Nested(K8SResourcesEntrySchema, allow_none=True)

    @staticmethod
    def schema_config():
        return K8SContainerResourcesConfig


class K8SContainerResourcesConfig(BaseConfig):
    """
    K8S container resources config.

    Args:
        limits: `K8SResourcesEntry`.
        requests: `K8SResourcesEntry`.
    """
    IDENTIFIER = 'resources'
    SCHEMA = K8SContainerResourcesSchema
    REDUCED_ATTRIBUTES = ['limits', 'requests']

    def __init__(self, limits=None, requests=None):
        self.limits = limits
        self.requests = requests


class K8SResourcesSchema(BaseSchema):
    limits = fields.Float(allow_none=True)
    requests = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return K8SResourcesConfig


class K8SResourcesConfig(BaseConfig):
    """
    K8S resources config.

    Args:
        limits: `float`.
        requests: `float`.
    """
    IDENTIFIER = 'resources'
    SCHEMA = K8SResourcesSchema
    REDUCED_ATTRIBUTES = ['limits', 'requests']

    def __init__(self, limits=None, requests=None):
        self.limits = limits
        self.requests = requests

    def __add__(self, other):
        if not other:
            return self

        if self.requests:
            if other.requests:
                self.requests += other.requests
        elif other.requests:
            self.requests = other.requests

        if self.limits:
            if other.limits:
                self.limits += other.limits
        elif other.limits:
            self.limits = other.limits

        return self


class PodResourcesSchema(BaseSchema):
    cpu = fields.Nested(K8SResourcesSchema, allow_none=True)
    memory = fields.Nested(K8SResourcesSchema, allow_none=True)
    gpu = fields.Nested(K8SResourcesSchema, allow_none=True)
    tpu = fields.Nested(K8SResourcesSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return PodResourcesConfig


class PodResourcesConfig(BaseConfig):
    """
    Pod resources config.

    Args:
        cpu: `K8SResourcesConfig`.
        memory: `K8SResourcesConfig`.
        gpu: `K8SResourcesConfig`.
        tpu: `K8SResourcesConfig`.
    """
    IDENTIFIER = 'pod_resources'
    SCHEMA = PodResourcesSchema
    REDUCED_ATTRIBUTES = ['cpu', 'memory', 'gpu', 'tpu']

    def __init__(self, cpu=None, memory=None, gpu=None, tpu=None):
        self.cpu = cpu
        self.memory = memory
        self.gpu = gpu
        self.tpu = tpu

    def __add__(self, other):  # pylint:disable=too-many-branches
        if not other:
            return self

        if self.cpu:
            if other.cpu:
                self.cpu += other.cpu
        elif other.cpu:
            self.cpu = K8SResourcesConfig()
            self.cpu += other.cpu

        if self.memory:
            if other.memory:
                self.memory += other.memory
        elif other.memory:
            self.memory = K8SResourcesConfig()
            self.memory += other.memory

        if self.gpu:
            if other.gpu:
                self.gpu += other.gpu
        elif other.gpu:
            self.gpu = K8SResourcesConfig()
            self.gpu += other.gpu

        if self.tpu:
            if other.tpu:
                self.tpu += other.tpu
        elif other.tpu:
            self.tpu = K8SResourcesConfig()
            self.tpu += other.tpu
        return self



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields


class UUID(fields.UUID):
    """A UUID field."""

    def _serialize(self, value, attr, obj, **kwargs):
        validated = str(self._validated(value).hex) if value is not None else None
        return super(fields.String, self)._serialize(validated, attr, obj)  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.group.early_stopping_policies import EarlyStoppingConfig
from polyaxon_schemas.ops.group.metrics import Optimization


class TestEarlyStoppingConfigs(TestCase):

    def test_early_stopping(self):
        config_dict = {
            'metric': 'loss',
            'value': 0.1,
        }
        config = EarlyStoppingConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict.pop('optimization') == Optimization.MAXIMIZE
        assert_equal_dict(config_to_dict, config_dict)

    def test_early_stopping_with_median_policy(self):
        config_dict = {
            'metric': 'loss',
            'value': 0.1,
            'optimization': Optimization.MINIMIZE,
            'policy': {'kind': 'median', 'evaluation_interval': 1}
        }
        config = EarlyStoppingConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_early_stopping_with_average_policy(self):
        config_dict = {
            'metric': 'loss',
            'value': 0.1,
            'optimization': Optimization.MINIMIZE,
            'policy': {'kind': 'average', 'evaluation_interval': 1}
        }
        config = EarlyStoppingConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_early_stopping_with_truncation_policy(self):
        config_dict = {
            'metric': 'loss',
            'value': 0.1,
            'optimization': Optimization.MAXIMIZE,
            'policy': {'kind': 'truncation', 'percent': 50, 'evaluation_interval': 1}
        }
        config = EarlyStoppingConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)



from hestia.unknown import UNKNOWN


class PodLifeCycle(object):
    CONTAINER_CREATING = 'ContainerCreating'
    PENDING = 'Pending'
    RUNNING = 'Running'
    SUCCEEDED = 'Succeeded'
    FAILED = 'Failed'
    UNKNOWN = UNKNOWN

    CHOICES = (
        (RUNNING, RUNNING),
        (PENDING, PENDING),
        (CONTAINER_CREATING, CONTAINER_CREATING),
        (SUCCEEDED, SUCCEEDED),
        (FAILED, FAILED),
    )

    DONE_STATUS = [FAILED, SUCCEEDED]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.environments.legacy import GPUOptionsConfig, SessionConfig, TFRunConfig
from polyaxon_schemas.ops.experiment.environment import TensorflowClusterConfig
from polyaxon_schemas.utils import TaskType


class TestLegacyConfigs(TestCase):

    def test_gpu_options_config(self):
        config_dict = {
            'gpu_memory_fraction': 0.8,
            'allow_growth': False,
            'per_process_gpu_memory_fraction': 0.4,
        }
        config = GPUOptionsConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_session_config(self):
        config_dict = {
            'log_device_placement': False,
            'allow_soft_placement': False,
            'intra_op_parallelism_threads': 2,
            'inter_op_parallelism_threads': 3,
            'gpu_options': GPUOptionsConfig().to_dict(),
        }
        config = SessionConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_indexed_session(self):
        config_dict = {
            'log_device_placement': False,
            'allow_soft_placement': False,
            'intra_op_parallelism_threads': 2,
            'inter_op_parallelism_threads': 3,
            'gpu_options': GPUOptionsConfig().to_dict(),
        }
        config = SessionConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_run_config(self):
        config_dict = {
            'tf_random_seed': 100,
            'save_summary_steps': 100,
            'save_checkpoints_secs': 600,
            'save_checkpoints_steps': None,
            'keep_checkpoint_max': 5,
            'keep_checkpoint_every_n_hours': 10000,
        }
        config = TFRunConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

        # Add session config
        config_dict['session'] = SessionConfig().to_dict()
        config = TFRunConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

        # Add cluster config
        config_dict['cluster'] = TensorflowClusterConfig(
            worker=[TaskType.WORKER], ps=[TaskType.PS]
        ).to_dict()
        config = TFRunConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema


class BaseEnvironmentSchema(BaseSchema):
    env_id = fields.Str()

    @staticmethod
    def schema_config():
        return BaseEnvironmentConfig


class BaseEnvironmentConfig(BaseConfig):
    def __init__(self, env_id):
        self.env_id = env_id


class GymEnvironmentSchema(BaseEnvironmentSchema):

    @staticmethod
    def schema_config():
        return GymEnvironmentConfig


class GymEnvironmentConfig(BaseEnvironmentConfig):
    IDENTIFIER = 'Gym'
    SCHEMA = GymEnvironmentSchema


class RegularizerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'environment'
    __configs__ = {
        GymEnvironmentConfig.IDENTIFIER: GymEnvironmentConfig,
    }
